unit cmplx80;
//-------------------------------------------------------------------------//
//                 Complex numbers routines library                        //
//               Copyright (c) 2001 by Serghei Gaivan                      //
//                    e-mail: gaivan@mail.hypermart.net                    //
//                  http://gaivan.hypermart.net                            //
//-------------------------------------------------------------------------//

interface

type
  complex = record
      x: extended;
      y: extended;
    end;

const
  ui: complex = (x: 0; y: -1);
  ur: complex = (x: 1; y:0);

function Re(z: complex): extended;          // z :--> Re(z)
function Im(z: complex): extended;          // z :--> Im(z)

//------ Unary operations ----------------------------------------------------//
function Conj(z: complex): complex;     // z :--> z*
function Neg(z: complex): complex;      // z :--> -z
function Flip(z: complex): complex;     // (x,y) :--> (y,x)
function RCW(z: complex): complex;      // (x,y) :-->  (-y,x)
function RCC(z: complex): complex;      // (x,y) :-->  (y,-x)

//------ Binary operations ---------------------------------------------------//
function CSum(z1,z2: complex): complex; // z1,z2 :--> z1+z2
function CSub(z1,z2: complex): complex; // z1,z2 :--> z1-z2
function CMul(z1,z2: complex): complex; // z1,z2 :--> z1*z2
function CDiv(z1,z2: complex): complex; // z1,z2 :--> z1/z2

//------ Standard routines ---------------------------------------------------//
function Polar(rho,phi: extended): complex; // (rho,phi) :--> z
function CAbs(z: complex): extended;        // z :--> |z|
function Arg(z: complex): extended;         // z :--> arg(z)
function Norm(z: complex): extended;        // z :--> |z|^2

//------ Algebraic functions -------------------------------------------------//
function CSqr(z: complex): complex;     // z :--> z^2
function CInv(z: complex): complex;     // z :--> 1/z
function CSqrt(z: complex): complex;    // z :--> Sqrt(z)
function Pow(z: complex; n: integer): complex;  // z :--> z^n

//------ Transcendent functions ----------------------------------------------//
function CLn(z: complex): complex;      // z :--> Ln(z)
function CExp(z: complex): complex;     // z :--> Exp(z)

//------ Trigonometric functions ---------------------------------------------//
function CSin(z: complex): complex;     // z :--> Sin(z)
function CCos(z: complex): complex;     // z :--> Cos(z)
function CTan(z: complex): complex;     // z :--> Tan(z)
function CCotan(z: complex): complex;   // z :--> Cotan(z)

//------ Hyperbolic functions ------------------------------------------------//
function CSinh(z: complex): complex;   // z :--> Sinh(z)
function CCosh(z: complex): complex;   // z :--> Cosh(z)
function CTanh(z: complex): complex;   // z :--> Tanh(z)
function CCotanh(z: complex): complex; // z :--> Cotanh(z)


implementation

//----------------------------------------------------------------------------//

function Re(z: complex): extended; register;
// z :--> Re(z)
asm
         FLD    TBYTE PTR [EAX]
end;

//----------------------------------------------------------------------------//

function Im(z: complex): extended; register;
// z :--> Im(z)
asm
         FLD    TBYTE PTR [EAX+$10]
end;

//----------------------------------------------------------------------------//
//------ Unary operations ----------------------------------------------------//
//----------------------------------------------------------------------------//

function Conj(z: complex): complex; register;
// z :--> z*
asm
         FLD    TBYTE PTR [EAX+$10]
         FCHS
         FSTP   TBYTE PTR [EDX+$10]
         FLD    TBYTE PTR [EAX]
         FSTP   TBYTE PTR [EDX]
end;

//----------------------------------------------------------------------------//

function Neg(z: complex): complex; register;
// (x,y) :--> (-x,-y)
asm
         FLD    TBYTE PTR [EAX]
         FCHS
         FSTP   TBYTE PTR [EDX]
         FLD    TBYTE PTR [EAX+$10]
         FCHS
         FSTP   TBYTE PTR [EDX+$10]
end;

//----------------------------------------------------------------------------//

function Flip(z: complex): complex;
// (x,y) :-->  (y,x)
asm
         FLD    TBYTE PTR [EAX+$10]
         FSTP   TBYTE PTR [EDX]
         FLD    TBYTE PTR [EAX]
         FSTP   TBYTE PTR [EDX+$10]
end;

//----------------------------------------------------------------------------//

function RCW(z: complex): complex; register;
// (x,y) :-->  (-y,x) that is z :--> i*z
asm
         FLD    TBYTE PTR [EAX+$10]
         FCHS
         FSTP   TBYTE PTR [EDX]
         FLD    TBYTE PTR [EAX]
         FSTP   TBYTE PTR [EDX+$10]
end;

//----------------------------------------------------------------------------//

function RCC(z: complex): complex; register;
// (x,y) :-->  (y,-x)
asm
         FLD    TBYTE PTR [EAX+$10]
         FSTP   TBYTE PTR [EDX]
         FLD    TBYTE PTR [EAX]
         FCHS
         FSTP   TBYTE PTR [EDX+$10]
end;

//----------------------------------------------------------------------------//
//------ Binary operations ---------------------------------------------------//
//----------------------------------------------------------------------------//

function CSum(z1,z2: complex): complex; register;
// z1,z2 :--> z1+z2
asm
         FLD      TBYTE PTR [EAX]
         FLD      TBYTE PTR [EDX]
         FADD
         FSTP     TBYTE PTR [ECX]
         FLD      TBYTE PTR [EAX+$10]
         FLD      TBYTE PTR [EDX+$10]
         FADD
         FSTP     TBYTE PTR [ECX+$10]
end;

//----------------------------------------------------------------------------//

function CSub(z1,z2: complex): complex; register;
// z1,z2 :--> z1-z2
asm
         FLD      TBYTE PTR [EAX]
         FLD      TBYTE PTR [EDX]
         FSUB
         FSTP     TBYTE PTR [ECX]
         FLD      TBYTE PTR [EAX+$10]
         FLD      TBYTE PTR [EDX+$10]
         FSUB
         FSTP     TBYTE PTR [ECX+$10]
end;

//----------------------------------------------------------------------------//

function CMul(z1,z2: complex): complex; register;
// z1,z2 :--> z1*z2
asm
         FLD      TBYTE PTR [EAX]
         FLD      TBYTE PTR [EDX]
         FLD      ST               // x2 x2 x1
         FMUL     ST,ST(2)         // x1*x2 x2 x1
         FLD      TBYTE PTR [EAX+$10]
         FXCH     ST(1)            // x1*x2 y1 x2 x1
         FLD      TBYTE PTR [EDX+$10]
         FXCH     ST(1)
         FLD      ST(1)
         FMUL     ST,ST(3)
         FSUB
         FSTP     TBYTE PTR [ECX]        // y2 y1 x2 x1
         FMULP    ST(3),ST(0)     //y1 x2 x1*y2
         FMUL                     // x2*y1 x1*y2
         FADD
         FSTP     TBYTE PTR [ECX+$10]
end;

//----------------------------------------------------------------------------//

function CDiv(z1,z2: complex): complex; register;
// z1,z2 :--> z1/z2
asm
         FLD      TBYTE PTR [EDX+$10]
         FLD      ST(0)
         FMUL     ST,ST
         FLD      TBYTE PTR [EDX]
         FXCH     ST(1)
         FLD      ST(1)
         FMUL     ST,ST
         FADD
         FLD1
         FDIVR
         FLD      TBYTE PTR [EAX]
         FLD      TBYTE PTR [EAX+$10]
         FXCH     ST(2)
         FLD      ST(1)
         FMUL     ST,ST(4)
         FLD      ST(3)
         FMUL     ST,ST(6)
         FADD
         FMUL     ST,ST(1)
         FSTP     TBYTE PTR [ECX]
         FXCH     ST(4)
         FMUL
         FXCH     ST(2)
         FMUL                  // x2*y1 x1*y2 1/norm
         FSUBR
         FMUL
         FSTP     TBYTE PTR [ECX+$10]
end;

//----------------------------------------------------------------------------//
//------ Standard routines ---------------------------------------------------//
//----------------------------------------------------------------------------//

function Polar(rho,phi: extended): complex; register;
// (rho,phi) :--> z
asm
         FLD      TBYTE PTR [EBP+$14]
         FLD      TBYTE PTR [EBP+$08]
         FSINCOS
         FMUL     ST,ST(2)
         FSTP     TBYTE PTR [EAX]
         FMUL
         FSTP     TBYTE PTR [EAX+$10]
end;

//----------------------------------------------------------------------------//

function CAbs(z: complex): extended; register;
// z :--> |z|
asm
         FLD      TBYTE PTR [EAX+$10]
         FMUL     ST,ST
         FLD      TBYTE PTR [EAX]
         FMUL     ST,ST
         FADD
         FSQRT
end;

//----------------------------------------------------------------------------//

function Arg(z: complex): extended; register;
// z :--> arg(z)
asm
         FLD    TBYTE PTR [EAX+$10]
         FLD    TBYTE PTR [EAX]
         FPATAN
end;

//----------------------------------------------------------------------------//

function Norm(z: complex): extended; register;
// z :--> |z|^2
asm
         FLD      TBYTE PTR [EAX+$10]
         FMUL     ST,ST
         FLD      TBYTE PTR [EAX]
         FMUL     ST,ST
         FADD
end;

//----------------------------------------------------------------------------//
//------ Algebraic functions -------------------------------------------------//
//----------------------------------------------------------------------------//

function CSqr(z: complex): complex; register;
// z :--> z^2
asm
         FLD      TBYTE PTR [EAX+$10]
         FLD      ST
         FMUL     ST,ST
         FLD      TBYTE PTR [EAX]
         FLD      ST
         FMUL     ST,ST
         FXCH     ST(3)
         FMUL
         FADD     ST,ST
         FSTP     TBYTE PTR [EDX+$10]
         FSUB
         FSTP     TBYTE PTR [EDX]
end;

//----------------------------------------------------------------------------//

function CSqrt(z: complex): complex; register;
// z :--> sqrt(z)
asm
         FLD      TBYTE PTR [EAX]
         FLD      ST
         FMUL     ST,ST
         FLD      TBYTE PTR [EAX+$10]
         FMUL     ST,ST
         FADD
         FSQRT
         FLD      ST(1)
         FADD     ST,ST(1)
         FABS
         FLD1
         FADD     ST,ST
         FDIV
         FSQRT
         FSTP     TBYTE PTR [EDX]
         FSUB
         FABS
         FLD1
         FADD     ST,ST
         FDIV
         FSQRT
         FSTP     TBYTE PTR [EDX+$10]
end;

//----------------------------------------------------------------------------//

function CInv(z: complex): complex; register;
// z :--> 1/z
asm
         FLD      TBYTE PTR [EAX+$10]
         FLD      ST
         FMUL     ST,ST
         FLD      TBYTE PTR [EAX]
         FXCH
         FLD      ST(1)
         FMUL     ST,ST
         FADD
         FLD1
         FDIVR
         FXCH     ST(2)
         FMUL     ST,ST(2)
         FSTP     TBYTE PTR [EDX+$10]
         FMUL
         FSTP     TBYTE PTR [EDX]
end;

//----------------------------------------------------------------------------//

function Pow(z: complex; n: integer): complex; register;
// z :--> z^n
asm
         FLD      TBYTE PTR [EAX]
         FLD      TBYTE PTR [EAX+$10]
         FLD1
         FLD      ST(2)
         FMUL     ST,ST
         FLD      ST(2)
         FMUL     ST,ST
         FADD
         FSQRT
         MOV      EAX,EDX
         JMP      @2
  @1:    FMUL     ST,ST
  @2:    SHR      EAX,1
         JNC      @1
         FMUL     ST(1),ST
         JNZ      @1
         FSTP     ST(0)
         FXCH     ST(2)
         FPATAN
         MOV      [ESP-$04],EDX
         FILD     DWORD PTR [ESP-$04]
         FMUL
         FSINCOS
         FMUL     ST,ST(2)
         FSTP     TBYTE PTR [ECX]
         FMUL
         FSTP     TBYTE PTR [ECX+$10]
end;

//----------------------------------------------------------------------------//
//------- Transcendent functions ---------------------------------------------//
//----------------------------------------------------------------------------//

function CLn(z: complex): complex; register;
// z :--> Ln(z)
asm
         FLD    TBYTE PTR [EAX+$10]
         FLD    TBYTE PTR [EAX]
         FLDLN2
         FLD1
         FADD     ST,ST
         FDIV
         FLD      ST(2)
         FMUL     ST,ST
         FLD      ST(2)
         FMUL     ST,ST
         FADD
         FYL2X
         FSTP     TBYTE PTR [EDX]
         FPATAN
         FSTP     TBYTE PTR [EDX+$10]
end;

//----------------------------------------------------------------------------//

function CExp(z: complex): complex; register;
// z :--> Exp(z)
asm
         FLD       TBYTE PTR [EAX]
         FLDL2E
         FMUL
         FLD       ST(0)
         FRNDINT
         FSUB      ST(1), ST
         FXCH      ST(1)
         F2XM1
         FLD1
         FADD
         FSCALE
         FSTP      ST(1)
         FLD       TBYTE PTR [EAX+$10]
         FSINCOS
         FMUL      ST,ST(2)
         FSTP      TBYTE PTR [EDX]
         FMUL
         FSTP      TBYTE PTR [EDX+$10]
end;

//----------------------------------------------------------------------------//
//------ Trigonometric functions ---------------------------------------------//
//----------------------------------------------------------------------------//

function CSin(z: complex): complex; register;
// z :--> Sin(z)
asm
         FLD       TBYTE PTR [EAX+$10]
         FLDL2E
         FMUL
         FLD       ST(0)
         FRNDINT
         FSUB      ST(1), ST
         FXCH      ST(1)
         F2XM1
         FLD1
         FADD
         FSCALE
         FSTP      ST(1)
         FLD1
         FLD       ST(1)
         FADD      ST,ST
         FDIV
         FXCH
         FLD1
         FADD      ST,ST
         FDIV
         FLD    TBYTE PTR [EAX]
         FSINCOS
         FLD       ST(2)
         FSUB      ST,ST(4)
         FMUL
         FSTP      TBYTE PTR [EDX+$10]
         FXCH      ST(2)
         FADD
         FMUL
         FSTP      TBYTE PTR [EDX]
end;

//----------------------------------------------------------------------------//

function CCos(z: complex): complex; register;
// z :--> Cos(z)
asm
         FLD       TBYTE PTR [EAX+$10]
         FLDL2E
         FMUL
         FLD       ST(0)
         FRNDINT
         FSUB      ST(1), ST
         FXCH      ST(1)
         F2XM1
         FLD1
         FADD
         FSCALE
         FSTP      ST(1)
         FLD1
         FLD       ST(1)
         FADD      ST,ST
         FDIV
         FXCH
         FLD1
         FADD      ST,ST
         FDIV
         FLD       TBYTE PTR [EAX]
         FSINCOS
         FLD       ST(2)
         FADD      ST,ST(4)
         FMUL
         FSTP      TBYTE PTR [EDX]
         FXCH      ST(2)
         FSUBR                     
         FMUL
         FSTP      TBYTE PTR [EDX+$10]
end;

//----------------------------------------------------------------------------//

function CTan(z: complex): complex; register;
// z :--> Tan(z)
asm
         FLD       TBYTE PTR [EAX]
         FADD      ST,ST
         FLD       TBYTE PTR [EAX+$10]
         FADD      ST,ST           // 2y 2x
         FLDL2E
         FMUL
         FLD       ST(0)
         FRNDINT
         FSUB      ST(1), ST
         FXCH      ST(1)
         F2XM1
         FLD1
         FADD
         FSCALE
         FSTP      ST(1)           // exp(2y) 2x
         FLD1                      // 1 exp(2y) 2x
         FDIV      ST(0),ST(1)     // exp(-2y) exp(2y) 2x
         FLD1
         FADD      ST,ST           // 2 exp(-2y) exp(2y) 2x
         FLD       ST(0)           // 2 2 exp(-2y) exp(2y) 2x
         FDIVP     ST(2),ST(0)     // 2 exp(-2y)/2 exp(2y) 2x
         FDIVP     ST(2),ST(0)     // exp(-2y)/2 exp(2y)/2 2x
         FLD       ST(1)           // exp(2y)/2 exp(-2y)/2 exp(2y)/2 2x
         FSUB      ST(0),ST(1)     // sinh(2y) exp(-2y)/2 exp(2y)/2 2x
         FXCH      ST(2)           // exp(2y)/2 exp(-2y)/2 sinh(2y) 2x
         FADD                      // cosh(2y) sinh(2y) 2x
         FXCH      ST(2)           // 2x sinh(2y) cosh(2y)
         FSINCOS                   // cos(2x) sin(2x) sinh(2y) cosh(2y)
         FADDP     ST(3),ST(0)     // sin(2x) sinh(2y) (cos+cosh)
         FDIV      ST(0),ST(2)
         FSTP      TBYTE PTR [EDX]        // sinh(2y) (cos+cosh)
         FDIVR
         FSTP      TBYTE PTR [EDX+$10]
end;

//----------------------------------------------------------------------------//

function CCotan(z: complex): complex; register;
// z :--> Cotan(z)
asm
         FLD       TBYTE PTR [EAX]
         FADD      ST,ST
         FLD       TBYTE PTR [EAX+$10]
         FADD      ST,ST           // 2y 2x
         FLDL2E
         FMUL
         FLD       ST(0)
         FRNDINT
         FSUB      ST(1), ST
         FXCH      ST(1)
         F2XM1
         FLD1
         FADD
         FSCALE
         FSTP      ST(1)           // exp(2y) 2x
         FLD1                      // 1 exp(2y) 2x
         FDIV      ST(0),ST(1)     // exp(-2y) exp(2y) 2x
         FLD1
         FADD      ST,ST           // 2 exp(-2y) exp(2y) 2x
         FLD       ST(0)           // 2 2 exp(-2y) exp(2y) 2x
         FDIVP     ST(2),ST(0)     // 2 exp(-2y)/2 exp(2y) 2x
         FDIVP     ST(2),ST(0)     // exp(-2y)/2 exp(2y)/2 2x
         FLD       ST(0)           // exp(-2y)/2 exp(-2y)/2 exp(2y)/2 2x
         FSUB      ST(0),ST(2)     // -sinh(2y) exp(-2y)/2 exp(2y)/2 2x
         FXCH      ST(2)
         FADD
         FXCH      ST(2)
         FSINCOS
         FSUBP     ST(3),ST(0)
         FDIV      ST(0),ST(2)
         FSTP      TBYTE PTR [EDX]
         FDIVR
         FSTP      TBYTE PTR [EDX+$10]
end;


//----------------------------------------------------------------------------//
//------ Hyperbolic functions -----------------------------------------------//
//----------------------------------------------------------------------------//

function CSinh(z: complex): complex; register;
// z :--> Sinh(z)
asm
         FLD    TBYTE PTR [EAX]
         FLDL2E
         FMUL
         FLD       ST(0)
         FRNDINT
         FSUB      ST(1), ST
         FXCH      ST(1)
         F2XM1
         FLD1
         FADD
         FSCALE
         FSTP      ST(1)           // exp(x)
         FLD1                      // 1 exp(x)
         FLD       ST(1)           // exp(x) 1 exp(x)
         FADD      ST,ST           // 2exp(x) 1 exp(x)
         FDIV                      // 1/2exp(x) exp(x)
         FXCH                      // exp(x) 1/2exp(x)
         FLD1                      // 1 exp(x) 1/2exp(x)
         FADD      ST,ST           // 2 exp(x) 1/2exp(x)
         FDIV                      // exp(x)/2 1/2exp(x)
         FLD       TBYTE PTR [EAX+$10]             // y tmp tmp2
         FSINCOS                   // cos(y) sin(y) tmp tmp2
         FLD       ST(2)           // tmp cos(y) sin(y) tmp tmp2
         FSUB      ST,ST(4)        // (tmp-tmp2) cos(y) sin(y) tmp tmp2
         FMUL
         FSTP      TBYTE PTR [EDX]        // sin(y) tmp tmp2
         FXCH      ST(2)           // tmp2 tmp sin(y)
         FADD                      // (tmp+tmp2 sin(y)
         FMUL
         FSTP      TBYTE PTR [EDX+$10]
end;

//----------------------------------------------------------------------------//

function CCosh(z: complex): complex; register;
// z :--> Cosh(z)
asm
         FLD    TBYTE PTR [EAX]
         FLDL2E
         FMUL
         FLD       ST(0)
         FRNDINT
         FSUB      ST(1), ST
         FXCH      ST(1)
         F2XM1
         FLD1
         FADD
         FSCALE
         FSTP      ST(1)           // exp(x)
         FLD1                      // 1 exp(x)
         FLD       ST(1)           // exp(x) 1 exp(x)
         FADD      ST,ST           // 2exp(x) 1 exp(x)
         FDIV                      // 1/2exp(x) exp(x)
         FXCH                      // exp(x) 1/2exp(x)
         FLD1                      // 1 exp(x) 1/2exp(x)
         FADD      ST,ST           // 2 exp(x) 1/2exp(x)
         FDIV                      // exp(x)/2 1/2exp(x)
         FLD       TBYTE PTR [EAX+$10]             // y tmp tmp2
         FSINCOS                   // cos(y) sin(y) tmp tmp2
         FLD       ST(2)           // tmp cos(y) sin(y) tmp tmp2
         FADD      ST,ST(4)        // (tmp+tmp2) cos(y) sin(y) tmp tmp2
         FMUL
         FSTP      TBYTE PTR [EDX]        // sin(y) tmp tmp2
         FXCH      ST(2)           // tmp2 tmp sin(y)
         FSUB                      // (tmp-tmp2 sin(y)
         FMUL
         FSTP      TBYTE PTR [EDX+$10]
end;

//----------------------------------------------------------------------------//

function CTanh(z: complex): complex; register;
// z :--> Tanh(z)
asm
         FLD       TBYTE PTR [EAX+$10]
         FADD      ST,ST
         FLD       TBYTE PTR [EAX]
         FADD      ST,ST           // 2x 2y
         FLDL2E
         FMUL
         FLD       ST(0)
         FRNDINT
         FSUB      ST(1), ST
         FXCH      ST(1)
         F2XM1
         FLD1
         FADD
         FSCALE
         FSTP      ST(1)           // exp(2x) 2y
         FLD1                      // 1 exp(2x) 2y
         FDIV      ST(0),ST(1)     // exp(-2x) exp(2x) 2y
         FLD1
         FADD      ST,ST           // 2 exp(-2x) exp(2x) 2y
         FLD       ST(0)           // 2 2 exp(-2x) exp(2x) 2y
         FDIVP     ST(2),ST(0)     // 2 exp(-2x)/2 exp(2x) 2y
         FDIVP     ST(2),ST(0)     // exp(-2x)/2 exp(2x)/2 2y
         FLD       ST(1)           // exp(2x)/2 exp(-2x)/2 exp(2x)/2 2y
         FSUB      ST(0),ST(1)     // sinh(2x) exp(-2x)/2 exp(2x)/2 2y
         FXCH      ST(2)           // exp(2x)/2 exp(-2x)/2 sinh(2x) 2y
         FADD                      // cosh(2x) sinh(2x) 2y
         FXCH      ST(2)           // 2y sinh(2x) cosh(2x)
         FSINCOS                   // cos(2y) sin(2y) sinh(2x) cosh(2x)
         FADDP     ST(3),ST(0)     // sin(2y) sinh(2x) (cos+cosh)
         FDIV      ST(0),ST(2)
         FSTP      TBYTE PTR [EDX+$10]        // sinh(2x) (cos+cosh)
         FDIVR
         FSTP      TBYTE PTR [EDX]
end;

//----------------------------------------------------------------------------//

function CCotanh(z: complex): complex; register;
// z :--> Cotanh(z)
asm
         FLD       TBYTE PTR [EAX+$10]
         FADD      ST,ST
         FLD       TBYTE PTR [EAX]
         FADD      ST,ST
         FLDL2E
         FMUL
         FLD       ST(0)
         FRNDINT
         FSUB      ST(1), ST
         FXCH      ST(1)
         F2XM1
         FLD1
         FADD
         FSCALE
         FSTP      ST(1)
         FLD1
         FDIV      ST(0),ST(1)
         FLD1
         FADD      ST,ST
         FLD       ST(0)
         FDIVP     ST(2),ST(0)
         FDIVP     ST(2),ST(0)
         FLD       ST(0)
         FSUB      ST(0),ST(2)
         FXCH      ST(2)
         FADD
         FXCH      ST(2)
         FSINCOS
         FSUBRP    ST(3),ST(0)
         FDIV      ST(0),ST(2)
         FSTP      TBYTE PTR [EDX+$10]
         FDIVR
         FSTP      TBYTE PTR [EDX]
end;

end. /// end of cmplx module ///
